/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.data.processor.common;

import com.google.common.collect.Iterables;
import com.google.common.util.concurrent.ListenableFuture;
import net.minecraft.init.Items;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.nbt.NBTUtil;
import net.minecraft.tileentity.TileEntitySkull;
import org.spongepowered.api.Sponge;
import org.spongepowered.api.profile.GameProfile;
import org.spongepowered.api.data.type.SkullType;
import org.spongepowered.api.data.type.SkullTypes;
import org.spongepowered.api.profile.GameProfileManager;
import org.spongepowered.common.SpongeImpl;
import org.spongepowered.common.data.manipulator.mutable.SpongeRepresentedPlayerData;
import org.spongepowered.common.data.util.NbtDataUtil;

import java.util.Optional;
import java.util.concurrent.ExecutionException;

import javax.annotation.Nullable;

public class SkullUtils {

    /**
     * There's not really a meaningful default value for this, since it's a CatalogType. However, the Vanilla give command defaults the skeleton type (index 0), so it's used as the default here.
     */
    public static final SkullType DEFAULT_TYPE = SkullTypes.SKELETON;

    public static boolean supportsObject(Object object) {
        return object instanceof TileEntitySkull || isValidItemStack(object);
    }

    public static SkullType getSkullType(int skullType) {
        return Iterables.get(SpongeImpl.getRegistry().getAllOf(SkullType.class), skullType);
    }

    public static boolean isValidItemStack(Object container) {
        return container instanceof ItemStack && ((ItemStack) container).func_77973_b().equals(Items.field_151144_bL);
    }

    public static SkullType getSkullType(TileEntitySkull tileEntitySkull) {
        return SkullUtils.getSkullType(tileEntitySkull.func_145904_a());
    }

    public static void setSkullType(TileEntitySkull tileEntitySkull, int skullType) {
        tileEntitySkull.func_152107_a(skullType);
        tileEntitySkull.func_70296_d();
        tileEntitySkull.func_145831_w().func_175689_h(tileEntitySkull.func_174877_v());
    }

    public static SkullType getSkullType(ItemStack itemStack) {
        return SkullUtils.getSkullType(itemStack.func_77960_j());
    }

    public static Optional<GameProfile> getProfile(TileEntitySkull entity) {
        return Optional.ofNullable((GameProfile) entity.func_152108_a());
    }

    public static boolean setProfile(TileEntitySkull skull, @Nullable GameProfile profile) {
        if (getSkullType(skull).equals(SkullTypes.PLAYER)) {
            final GameProfile newProfile = SpongeRepresentedPlayerData.NULL_PROFILE.equals(profile) ? null : resolveProfileIfNecessary(profile);
            skull.func_152106_a((com.mojang.authlib.GameProfile) newProfile);
            skull.func_70296_d();
            skull.func_145831_w().func_175689_h(skull.func_174877_v());
            return true;
        } else {
            return false;
        }
    }

    public static Optional<GameProfile> getProfile(ItemStack skull) {
        if (isValidItemStack(skull) && getSkullType(skull).equals(SkullTypes.PLAYER)) {
            final NBTTagCompound nbt = skull.func_179543_a(NbtDataUtil.ITEM_SKULL_OWNER, false);
            final com.mojang.authlib.GameProfile mcProfile = nbt == null ? null : NBTUtil.func_152459_a(nbt);
            return Optional.ofNullable((GameProfile) mcProfile);
        }
        return Optional.empty();
    }

    public static boolean setProfile(ItemStack skull, @Nullable GameProfile profile) {
        if (isValidItemStack(skull) && getSkullType(skull).equals(SkullTypes.PLAYER)) {
            if (profile == null || profile.equals(SpongeRepresentedPlayerData.NULL_PROFILE)) {
                if (skull.func_77978_p() != null) {
                    skull.func_77978_p().func_82580_o(NbtDataUtil.ITEM_SKULL_OWNER);
                }
            } else {
                final NBTTagCompound nbt = new NBTTagCompound();
                NBTUtil.func_180708_a(nbt, (com.mojang.authlib.GameProfile) resolveProfileIfNecessary(profile));
                skull.func_77983_a(NbtDataUtil.ITEM_SKULL_OWNER, nbt);
            }
            return true;
        }
        return false;
    }

    private static @Nullable GameProfile resolveProfileIfNecessary(@Nullable GameProfile profile) {
        if (profile == null) {
            return null;
        }
        // Skulls need a name in order to properly display -> resolve if no name is contained in the given profile
        final GameProfileManager resolver = Sponge.getGame().getServer().getGameProfileManager();
        if (profile.getName() == null || profile.getName().isEmpty()) {
            final ListenableFuture<GameProfile> future = resolver.get(profile.getUniqueId());
            try {
                return future.get();
            } catch (InterruptedException | ExecutionException e) {
                SpongeImpl.getLogger().debug("Exception while trying to resolve GameProfile: ", e);
                return null;
            }
        } else if (profile.getUniqueId() == null) {
            final ListenableFuture<GameProfile> future = resolver.get(profile.getName());
            try {
                return future.get();
            } catch (InterruptedException | ExecutionException e) {
                SpongeImpl.getLogger().debug("Exception while trying to resolve GameProfile: ", e);
                return null;
            }
        } else {
            return profile;
        }
    }

}
