/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

package org.spongepowered.common.entity.living.human;

import com.google.common.cache.CacheBuilder;
import com.google.common.cache.CacheLoader;
import com.google.common.cache.LoadingCache;
import com.google.common.collect.Maps;
import com.mojang.authlib.GameProfile;
import com.mojang.authlib.properties.PropertyMap;
import net.minecraft.enchantment.EnchantmentHelper;
import net.minecraft.entity.Entity;
import net.minecraft.entity.EntityCreature;
import net.minecraft.entity.EntityLivingBase;
import net.minecraft.entity.SharedMonsterAttributes;
import net.minecraft.entity.player.EntityPlayerMP;
import net.minecraft.item.Item;
import net.minecraft.item.ItemStack;
import net.minecraft.nbt.NBTTagCompound;
import net.minecraft.network.Packet;
import net.minecraft.network.play.server.S0CPacketSpawnPlayer;
import net.minecraft.network.play.server.S13PacketDestroyEntities;
import net.minecraft.network.play.server.S38PacketPlayerListItem;
import net.minecraft.scoreboard.Team;
import net.minecraft.server.MinecraftServer;
import net.minecraft.util.DamageSource;
import net.minecraft.util.MathHelper;
import net.minecraft.world.World;
import net.minecraft.world.WorldSettings;
import org.spongepowered.api.data.DataTransactionResult;
import org.spongepowered.api.data.key.Keys;
import org.spongepowered.api.data.value.immutable.ImmutableValue;
import org.spongepowered.api.entity.ArmorEquipable;
import org.spongepowered.api.scoreboard.TeamMember;
import org.spongepowered.api.text.Text;
import org.spongepowered.asm.mixin.Overwrite;
import org.spongepowered.common.SpongeImpl;
import org.spongepowered.common.data.value.immutable.ImmutableSpongeValue;
import org.spongepowered.common.interfaces.entity.IMixinEntity;

import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.concurrent.TimeUnit;

import net.minecraft.network.play.server.S38PacketPlayerListItem.AddPlayerData;
/*
 * Notes
 *
 * The label above the player's head is always visible unless the human is in
 * a team with invisible labels set to true. Could we leverage this at all?
 *
 * Hostile mobs don't attack the human, should this be default behaviour?
 */
public class EntityHuman extends EntityCreature implements TeamMember {

    // According to http://wiki.vg/Mojang_API#UUID_-.3E_Profile_.2B_Skin.2FCape
    // you can access this data once per minute, lets cache for 2 minutes
    private static final LoadingCache<UUID, PropertyMap> PROPERTIES_CACHE = CacheBuilder.newBuilder().expireAfterWrite(2, TimeUnit.MINUTES)
            .build(new CacheLoader<UUID, PropertyMap>() {

                @Override
                public PropertyMap load(UUID uuid) throws Exception {
                    return MinecraftServer.func_71276_C().func_147130_as().fillProfileProperties(new GameProfile(uuid, ""), true)
                            .getProperties();
                }
            });

    // A queue of packets waiting to send to players tracking this human
    private final Map<UUID, List<Packet[]>> playerPacketMap = Maps.newHashMap();

    private GameProfile fakeProfile;
    private UUID skinUuid;

    public EntityHuman(World worldIn) {
        super(worldIn);
        this.fakeProfile = new GameProfile(this.field_96093_i, "");
        this.func_70105_a(0.6F, 1.8F);
        this.func_98053_h(true);
    }

    @Override
    protected void func_110147_ax() {
        super.func_110147_ax();
        this.func_110140_aT().func_111150_b(SharedMonsterAttributes.field_111264_e).func_111128_a(1.0D);
    }

    @Override
    protected void func_70088_a() {
        super.func_70088_a();
        this.field_70180_af.func_75682_a(16, (byte) 0);
        this.field_70180_af.func_75682_a(17, 0.0F);
        this.field_70180_af.func_75682_a(18, 0);
        // Enables all skin features
        this.field_70180_af.func_75682_a(10, (byte) 0xFF);
    }

    @Override
    public Text getTeamRepresentation() {
        return Text.of(this.fakeProfile.getName());
    }

    @Override
    public Team func_96124_cp() {
        return this.field_70170_p.func_96441_U().func_96509_i(this.fakeProfile.getName());
    }

    @Override
    public void func_96094_a(String name) {
        if (name.length() > 16) {
            // Vanilla restriction
            name = name.substring(0, 16);
        }
        if (this.func_95999_t().equals(name)) {
            return;
        }
        super.func_96094_a(name);
        this.renameProfile(name);
        if (this.isAliveAndInWorld()) {
            this.respawnOnClient();
        }
    }

    @Override
    public void func_70037_a(NBTTagCompound tagCompund) {
        super.func_70037_a(tagCompund);
        String skinUuidString = ((IMixinEntity) this).getSpongeData().func_74779_i("skinUuid");
        if (!skinUuidString.isEmpty()) {
            this.updateFakeProfileWithSkin(UUID.fromString(skinUuidString));
        }
    }

    @Override
    public void func_70014_b(NBTTagCompound tagCompound) {
        super.func_70014_b(tagCompound);
        NBTTagCompound spongeData = ((IMixinEntity) this).getSpongeData();
        if (this.skinUuid != null) {
            spongeData.func_74778_a("skinUuid", this.skinUuid.toString());
        } else {
            spongeData.func_82580_o("skinUuid");
        }
    }

    @Override
    public void func_70636_d() {
        super.func_70636_d();
        this.func_82168_bl();
    }

    @Override
    public int func_82145_z() {
        return 80;
    }

    @Override
    protected String func_145776_H() {
        return "game.player.swim";
    }

    @Override
    protected String func_145777_O() {
        return "game.player.swim.splash";
    }

    @Override
    public int func_82147_ab() {
        return 10;
    }

    @Override
    public void func_70645_a(DamageSource cause) {
        super.func_70645_a(cause);
        this.func_70105_a(0.2F, 0.2F);
        this.func_70107_b(this.field_70165_t, this.field_70163_u, this.field_70161_v);
        this.field_70181_x = 0.1D;
        if (cause != null) {
            this.field_70159_w = (double) (-MathHelper.func_76134_b((this.field_70739_aP + this.field_70177_z) * (float) Math.PI / 180.0F) * 0.1F);
            this.field_70179_y = (double) (-MathHelper.func_76126_a((this.field_70739_aP + this.field_70177_z) * (float) Math.PI / 180.0F) * 0.1F);
        } else {
            this.field_70159_w = this.field_70179_y = 0.0D;
        }
    }

    @Override
    protected String func_70621_aR() {
        return "game.player.hurt";
    }

    @Override
    protected String func_70673_aS() {
        return "game.player.die";
    }

    @Override
    public double func_70033_W() {
        return -0.35D;
    }

    @Override
    public float func_70689_ay() {
        return (float) this.func_110148_a(SharedMonsterAttributes.field_111263_d).func_111126_e();
    }

    @Override
    protected String func_146067_o(int damageValue) {
        return damageValue > 4 ? "game.player.hurt.fall.big" : "game.player.hurt.fall.small";
    }

    @Override
    public float func_70047_e() {
        return 1.62f;
    }

    @Override
    public float func_110139_bj() {
        return this.func_70096_w().func_111145_d(17);
    }

    @Override
    public void func_110149_m(float amount) {
        if (amount < 0.0F) {
            amount = 0.0F;
        }
        this.func_70096_w().func_75692_b(17, amount);
    }

    @Override
    protected float func_110146_f(float p_110146_1_, float p_110146_2_) {
        float retValue = super.func_110146_f(p_110146_1_, p_110146_2_);
        // Make the body rotation follow head rotation
        this.field_70177_z = this.field_70759_as;
        return retValue;
    }

    @Override
    public boolean func_70652_k(Entity entityIn) {
        super.func_70652_k(entityIn);
        func_71038_i();
        float f = (float) this.func_110148_a(SharedMonsterAttributes.field_111264_e).func_111126_e();
        int i = 0;

        if (entityIn instanceof EntityLivingBase) {
            f += EnchantmentHelper.func_152377_a(this.func_70694_bm(), ((EntityLivingBase) entityIn).func_70668_bt());
            i += EnchantmentHelper.func_77501_a(this);
        }

        boolean flag = entityIn.func_70097_a(DamageSource.func_76358_a(this), f);

        if (flag) {
            if (i > 0) {
                entityIn.func_70024_g((double) (-MathHelper.func_76126_a(this.field_70177_z * (float) Math.PI / 180.0F) * (float) i * 0.5F), 0.1D,
                        (double) (MathHelper.func_76134_b(this.field_70177_z * (float) Math.PI / 180.0F) * (float) i * 0.5F));
                this.field_70159_w *= 0.6D;
                this.field_70179_y *= 0.6D;
            }

            int j = EnchantmentHelper.func_90036_a(this);

            if (j > 0) {
                entityIn.func_70015_d(j * 4);
            }

            this.func_174815_a(this, entityIn);
        }

        return flag;
    }

    private void renameProfile(String newName) {
        PropertyMap props = this.fakeProfile.getProperties();
        this.fakeProfile = new GameProfile(this.fakeProfile.getId(), newName);
        this.fakeProfile.getProperties().putAll(props);
    }

    private boolean updateFakeProfileWithSkin(UUID skin) {
        PropertyMap properties = PROPERTIES_CACHE.getUnchecked(skin);
        if (properties.isEmpty()) {
            return false;
        }
        this.fakeProfile.getProperties().replaceValues("textures", properties.get("textures"));
        this.skinUuid = skin;
        return true;
    }

    public void removeFromTabListDelayed(EntityPlayerMP player, S38PacketPlayerListItem removePacket) {
        int delay = SpongeImpl.getGlobalConfig().getConfig().getEntity().getHumanPlayerListRemoveDelay();
        Runnable removeTask = () -> this.pushPackets(player, removePacket);
        if (delay == 0) {
            removeTask.run();
        } else {
            SpongeImpl.getGame().getScheduler().createTaskBuilder().execute(removeTask).delayTicks(delay).submit(SpongeImpl.getPlugin());
        }
    }

    public boolean setSkinUuid(UUID skin) {
        if (!MinecraftServer.func_71276_C().func_71266_T()) {
            // Skins only work when online-mode = true
            return false;
        }
        if (skin.equals(this.skinUuid)) {
            return true;
        }
        if (!updateFakeProfileWithSkin(skin)) {
            return false;
        }
        if (this.isAliveAndInWorld()) {
            this.respawnOnClient();
        }
        return true;
    }

    public UUID getSkinUuid() {
        return this.skinUuid;
    }

    public DataTransactionResult removeSkin() {
        if (this.skinUuid == null) {
            return DataTransactionResult.successNoData();
        }
        this.fakeProfile.getProperties().removeAll("textures");
        ImmutableValue<?> oldValue = new ImmutableSpongeValue<>(Keys.SKIN_UNIQUE_ID, this.skinUuid);
        this.skinUuid = null;
        if (this.isAliveAndInWorld()) {
            this.respawnOnClient();
        }
        return DataTransactionResult.builder().result(DataTransactionResult.Type.SUCCESS).replace(oldValue).build();
    }

    private boolean isAliveAndInWorld() {
        return this.field_70170_p.func_73045_a(this.func_145782_y()) == this && !this.field_70128_L;
    }

    private void respawnOnClient() {
        this.pushPackets(new S13PacketDestroyEntities(this.func_145782_y()), this.createPlayerListPacket(S38PacketPlayerListItem.Action.ADD_PLAYER));
        this.pushPackets(this.createSpawnPacket());
        removeFromTabListDelayed(null, this.createPlayerListPacket(S38PacketPlayerListItem.Action.REMOVE_PLAYER));
    }

    /**
     * Can the fake profile be removed from the tab list immediately (i.e. as
     * soon as the human has spawned).
     *
     * @return Whether it can be removed with 0 ticks delay
     */
    public boolean canRemoveFromListImmediately() {
        return !this.fakeProfile.getProperties().containsKey("textures");
    }

    /**
     * Called when a player stops tracking this human.
     *
     * Removes the player from the packet queue and sends them a REMOVE_PLAYER
     * tab list packet to make sure the human is not on it.
     *
     * @param player The player that has stopped tracking this human
     */
    public void onRemovedFrom(EntityPlayerMP player) {
        this.playerPacketMap.remove(player.func_110124_au());
        player.field_71135_a.func_147359_a(this.createPlayerListPacket(S38PacketPlayerListItem.Action.REMOVE_PLAYER));
    }

    /**
     * Creates a {@link S0CPacketSpawnPlayer} packet.
     *
     * Copied directly from the constructor of the packet, because that can't be
     * used as we're not an EntityPlayer.
     *
     * @return A new spawn packet
     */
    public S0CPacketSpawnPlayer createSpawnPacket() {
        S0CPacketSpawnPlayer packet = new S0CPacketSpawnPlayer();
        packet.field_148957_a = this.func_145782_y();
        packet.field_179820_b = this.fakeProfile.getId();
        packet.field_148956_c = MathHelper.func_76128_c(this.field_70165_t * 32.0D);
        packet.field_148953_d = MathHelper.func_76128_c(this.field_70163_u * 32.0D);
        packet.field_148954_e = MathHelper.func_76128_c(this.field_70161_v * 32.0D);
        packet.field_148951_f = (byte) ((int) (this.field_70177_z * 256.0F / 360.0F));
        packet.field_148952_g = (byte) ((int) (this.field_70125_A * 256.0F / 360.0F));
        ItemStack itemstack = (ItemStack) ((ArmorEquipable) this).getItemInHand().orElse(null);
        packet.field_148959_h = itemstack == null ? 0 : Item.func_150891_b(itemstack.func_77973_b());
        packet.field_148960_i = this.func_70096_w();
        return packet;
    }

    /**
     * Creates a {@link S38PacketPlayerListItem} packet with the given action.
     *
     * @param action The action to apply on the tab list
     * @return A new tab list packet
     */
    @SuppressWarnings("unchecked")
    public S38PacketPlayerListItem createPlayerListPacket(S38PacketPlayerListItem.Action action) {
        S38PacketPlayerListItem packet = new S38PacketPlayerListItem(action);
        packet.field_179769_b.add(packet.new AddPlayerData(this.fakeProfile, 0, WorldSettings.GameType.NOT_SET, this.func_145748_c_()));
        return packet;
    }

    /**
     * Push the given packets to all players tracking this human.
     *
     * @param packets All packets to send in a single tick
     */
    public void pushPackets(Packet... packets) {
        this.pushPackets(null, packets); // null = all players
    }

    /**
     * Push the given packets to the given player (who must be tracking this
     * human).
     *
     * @param player The player tracking this human
     * @param packets All packets to send in a single tick
     */
    public void pushPackets(EntityPlayerMP player, Packet... packets) {
        List<Packet[]> queue = this.playerPacketMap.get(player);
        if (queue == null) {
            this.playerPacketMap.put(player == null ? null : player.func_110124_au(), queue = new ArrayList<>());
        }
        queue.add(packets);
    }

    /**
     * (Internal) Pops the packets off the queue for the given player.
     *
     * @param player The player to get packets for (or null for all players)
     * @return An array of packets to send in a single tick
     */
    public Packet[] popQueuedPackets(EntityPlayerMP player) {
        List<Packet[]> queue = this.playerPacketMap.get(player == null ? null : player.func_110124_au());
        return queue == null || queue.isEmpty() ? null : queue.remove(0);
    }
}
