/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.mixin.core.world.biome;

import net.minecraft.block.BlockStone;
import net.minecraft.block.state.IBlockState;
import net.minecraft.init.Blocks;
import net.minecraft.world.World;
import net.minecraft.world.biome.BiomeDecorator;
import net.minecraft.world.biome.BiomeGenBase;
import net.minecraft.world.biome.BiomeGenMutated;
import net.minecraft.world.gen.ChunkProviderSettings;
import org.spongepowered.api.block.BlockState;
import org.spongepowered.api.data.type.PlantTypes;
import org.spongepowered.api.data.type.ShrubTypes;
import org.spongepowered.api.util.annotation.NonnullByDefault;
import org.spongepowered.api.util.weighted.ChanceTable;
import org.spongepowered.api.util.weighted.EmptyObject;
import org.spongepowered.api.util.weighted.SeededVariableAmount;
import org.spongepowered.api.util.weighted.VariableAmount;
import org.spongepowered.api.util.weighted.WeightedObject;
import org.spongepowered.api.world.biome.BiomeGenerationSettings;
import org.spongepowered.api.world.biome.BiomeType;
import org.spongepowered.api.world.biome.GroundCoverLayer;
import org.spongepowered.api.world.gen.populator.BigMushroom;
import org.spongepowered.api.world.gen.populator.Cactus;
import org.spongepowered.api.world.gen.populator.DeadBush;
import org.spongepowered.api.world.gen.populator.Flower;
import org.spongepowered.api.world.gen.populator.Forest;
import org.spongepowered.api.world.gen.populator.Mushroom;
import org.spongepowered.api.world.gen.populator.Ore;
import org.spongepowered.api.world.gen.populator.Pumpkin;
import org.spongepowered.api.world.gen.populator.RandomBlock;
import org.spongepowered.api.world.gen.populator.Reed;
import org.spongepowered.api.world.gen.populator.SeaFloor;
import org.spongepowered.api.world.gen.populator.Shrub;
import org.spongepowered.api.world.gen.populator.WaterLily;
import org.spongepowered.api.world.gen.type.BiomeTreeTypes;
import org.spongepowered.api.world.gen.type.MushroomType;
import org.spongepowered.api.world.gen.type.MushroomTypes;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.common.interfaces.world.biome.IBiomeGenBase;
import org.spongepowered.common.world.biome.SpongeBiomeGenerationSettings;
import org.spongepowered.common.world.gen.WorldGenConstants;
import org.spongepowered.common.world.gen.populators.WrappedBiomeDecorator;

@NonnullByDefault
@Mixin(BiomeGenBase.class)
public abstract class MixinBiomeGenBase implements BiomeType, IBiomeGenBase {

    @Shadow public String biomeName;
    @Shadow public float temperature;
    @Shadow public float rainfall;
    @Shadow public float minHeight;
    @Shadow public float maxHeight;
    @Shadow public IBlockState topBlock;
    @Shadow public IBlockState fillerBlock;
    @Shadow public BiomeDecorator theBiomeDecorator;

    @Override
    public BiomeGenerationSettings initPopulators(World world) {
        SpongeBiomeGenerationSettings gensettings = new SpongeBiomeGenerationSettings();
        gensettings.getPopulators().clear();
        gensettings.getGenerationPopulators().clear();
        gensettings.getGroundCoverLayers().clear();
        buildPopulators(world, gensettings);
        if (!getClass().getName().startsWith("net.minecraft")) {
            gensettings.getPopulators().add(new WrappedBiomeDecorator((BiomeGenBase) (Object) this));
        } else if (!this.theBiomeDecorator.getClass().getName().startsWith("net.minecraft")) {
            gensettings.getPopulators().add(new WrappedBiomeDecorator(this.theBiomeDecorator));
        }
        return gensettings;
    }

    @Override
    public void buildPopulators(World world, SpongeBiomeGenerationSettings gensettings) {
        BiomeDecorator theBiomeDecorator = this.theBiomeDecorator;
        if (BiomeGenMutated.class.isAssignableFrom(this.getClass())) {
            theBiomeDecorator = ((BiomeGenMutated) (Object) this).field_150611_aD.field_76760_I;
        }

        gensettings.getGroundCoverLayers().add(new GroundCoverLayer((BlockState) this.topBlock, SeededVariableAmount.fixed(1)));
        gensettings.getGroundCoverLayers().add(new GroundCoverLayer((BlockState) this.fillerBlock, WorldGenConstants.GROUND_COVER_DEPTH));

        String s = world.func_72912_H().func_82571_y();
        ChunkProviderSettings settings;
        if (s != null) {
            settings = ChunkProviderSettings.Factory.func_177865_a(s).func_177864_b();
        } else {
            settings = ChunkProviderSettings.Factory.func_177865_a("").func_177864_b();
        }

        Ore dirt = Ore.builder()
                .ore((BlockState) Blocks.field_150346_d.func_176223_P())
                .size(settings.field_177789_I)
                .perChunk(settings.field_177790_J)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177791_K, settings.field_177784_L))
                .build();
        gensettings.getPopulators().add(dirt);

        Ore gravel = Ore.builder()
                .ore((BlockState) Blocks.field_150351_n.func_176223_P())
                .size(settings.field_177785_M)
                .perChunk(settings.field_177786_N)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177787_O, settings.field_177797_P))
                .build();
        gensettings.getPopulators().add(gravel);

        Ore diorite = Ore.builder()
                .ore((BlockState) Blocks.field_150348_b.func_176223_P().func_177226_a(BlockStone.field_176247_a, BlockStone.EnumType.DIORITE))
                .size(settings.field_177792_U)
                .perChunk(settings.field_177795_V)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177794_W, settings.field_177801_X))
                .build();
        gensettings.getPopulators().add(diorite);

        Ore granite = Ore.builder()
                .ore((BlockState) Blocks.field_150348_b.func_176223_P().func_177226_a(BlockStone.field_176247_a, BlockStone.EnumType.GRANITE))
                .size(settings.field_177796_Q)
                .perChunk(settings.field_177799_R)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177798_S, settings.field_177793_T))
                .build();
        gensettings.getPopulators().add(granite);

        Ore andesite = Ore.builder()
                .ore((BlockState) Blocks.field_150348_b.func_176223_P().func_177226_a(BlockStone.field_176247_a, BlockStone.EnumType.ANDESITE))
                .size(settings.field_177800_Y)
                .perChunk(settings.field_177802_Z)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177846_aa, settings.field_177847_ab))
                .build();
        gensettings.getPopulators().add(andesite);

        Ore coal = Ore.builder()
                .ore((BlockState) Blocks.field_150365_q.func_176223_P())
                .size(settings.field_177844_ac)
                .perChunk(settings.field_177845_ad)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177851_ae, settings.field_177853_af))
                .build();
        gensettings.getPopulators().add(coal);

        Ore iron = Ore.builder()
                .ore((BlockState) Blocks.field_150366_p.func_176223_P())
                .size(settings.field_177848_ag)
                .perChunk(settings.field_177849_ah)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177832_ai, settings.field_177834_aj))
                .build();
        gensettings.getPopulators().add(iron);

        Ore gold = Ore.builder()
                .ore((BlockState) Blocks.field_150352_o.func_176223_P())
                .size(settings.field_177828_ak)
                .perChunk(settings.field_177830_al)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177840_am, settings.field_177842_an))
                .build();
        gensettings.getPopulators().add(gold);

        Ore redstone = Ore.builder()
                .ore((BlockState) Blocks.field_150450_ax.func_176223_P())
                .size(settings.field_177836_ao)
                .perChunk(settings.field_177838_ap)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177818_aq, settings.field_177816_ar))
                .build();
        gensettings.getPopulators().add(redstone);

        Ore diamond = Ore.builder()
                .ore((BlockState) Blocks.field_150482_ag.func_176223_P())
                .size(settings.field_177814_as)
                .perChunk(settings.field_177812_at)
                .height(VariableAmount.baseWithRandomAddition(settings.field_177826_au, settings.field_177824_av))
                .build();
        gensettings.getPopulators().add(diamond);

        Ore lapis = Ore.builder()
                .ore((BlockState) Blocks.field_150369_x.func_176223_P())
                .size(settings.field_177822_aw)
                .perChunk(settings.field_177820_ax)
                .height(VariableAmount.baseWithVariance(settings.field_177807_ay, settings.field_177805_az))
                .build();
        gensettings.getPopulators().add(lapis);

        if (theBiomeDecorator.field_76805_H > 0) {
            SeaFloor sand = SeaFloor.builder()
                    .block((BlockState) Blocks.field_150354_m.func_176223_P())
                    .radius(VariableAmount.baseWithRandomAddition(2, 5))
                    .depth(2)
                    .perChunk(theBiomeDecorator.field_76805_H)
                    .replace(WorldGenConstants.DIRT_OR_GRASS)
                    .build();
            gensettings.getPopulators().add(sand);
        }
        if (theBiomeDecorator.field_76806_I > 0) {
            SeaFloor clay = SeaFloor.builder()
                    .block((BlockState) Blocks.field_150435_aG.func_176223_P())
                    .radius(VariableAmount.baseWithRandomAddition(2, 2))
                    .depth(1)
                    .perChunk(theBiomeDecorator.field_76806_I)
                    .replace(WorldGenConstants.DIRT)
                    .build();
            gensettings.getPopulators().add(clay);
        }
        if (theBiomeDecorator.field_76801_G > 0) {
            SeaFloor gravelSeaFloor = SeaFloor.builder()
                    .block((BlockState) Blocks.field_150351_n.func_176223_P())
                    .radius(VariableAmount.baseWithRandomAddition(2, 4))
                    .depth(2)
                    .perChunk(theBiomeDecorator.field_76801_G)
                    .replace(WorldGenConstants.DIRT_OR_GRASS)
                    .build();
            gensettings.getPopulators().add(gravelSeaFloor);
        }
        Forest forest = Forest.builder()
                .type(BiomeTreeTypes.OAK.getPopulatorObject(), 9)
                .type(BiomeTreeTypes.OAK.getLargePopulatorObject().get(), 1)
                .perChunk(VariableAmount.baseWithOptionalAddition(theBiomeDecorator.field_76832_z, 1, 0.1))
                .build();
        gensettings.getPopulators().add(forest);

        if (theBiomeDecorator.field_76807_J > 0) {
            BigMushroom mushroom = BigMushroom.builder()
                    .mushroomsPerChunk(theBiomeDecorator.field_76807_J)
                    .type(MushroomTypes.BROWN.getPopulatorObject(), 1)
                    .type(MushroomTypes.RED.getPopulatorObject(), 1)
                    .build();
            gensettings.getPopulators().add(mushroom);
        }
        if (theBiomeDecorator.field_76802_A > 0) {
            Flower flower = Flower.builder()
                    .perChunk(theBiomeDecorator.field_76802_A * 64)
                    .type(PlantTypes.DANDELION, 2)
                    .type(PlantTypes.POPPY, 1)
                    .build();
            gensettings.getPopulators().add(flower);
        }
        if (theBiomeDecorator.field_76803_B > 0) {
            Shrub grass = Shrub.builder()
                    .perChunk(theBiomeDecorator.field_76803_B * 128)
                    .type(ShrubTypes.TALL_GRASS, 1)
                    .build();
            gensettings.getPopulators().add(grass);
        }
        if (theBiomeDecorator.field_76804_C > 0) {
            DeadBush deadBush = DeadBush.builder()
                    .perChunk(theBiomeDecorator.field_76804_C)
                    .build();
            gensettings.getPopulators().add(deadBush);
        }
        if (theBiomeDecorator.field_76833_y > 0) {
            WaterLily waterLily = WaterLily.builder()
                    .perChunk(theBiomeDecorator.field_76833_y * 10)
                    .build();
            gensettings.getPopulators().add(waterLily);
        }
        ChanceTable<MushroomType> types = new ChanceTable<MushroomType>();
        types.add(new WeightedObject<MushroomType>(MushroomTypes.BROWN, 2));
        types.add(new WeightedObject<MushroomType>(MushroomTypes.RED, 1));
        types.add(new EmptyObject<>(5));
        Mushroom smallMushroom = Mushroom.builder()
                .types(types)
                .mushroomsPerChunk(theBiomeDecorator.field_76798_D + 1)
                .build();
        gensettings.getPopulators().add(smallMushroom);
        Reed reed = Reed.builder()
                .perChunk(theBiomeDecorator.field_76799_E + 10)
                .reedHeight(VariableAmount.baseWithRandomAddition(2, VariableAmount.baseWithRandomAddition(1, 3)))
                .build();
        gensettings.getPopulators().add(reed);
        Pumpkin pumpkin = Pumpkin.builder()
                .perChunk(64)
                .chance(1 / 32d)
                .build();
        gensettings.getPopulators().add(pumpkin);
        if (theBiomeDecorator.field_76800_F > 0) {
            Cactus cactus = Cactus.builder()
                    .cactiPerChunk(VariableAmount.baseWithOptionalAddition(0,
                            VariableAmount.baseWithRandomAddition(1, VariableAmount.baseWithOptionalAddition(2, 3, 0.5)), 0.8))
                    .build();
            gensettings.getPopulators().add(cactus);
        }
        if (theBiomeDecorator.field_76808_K) {
            RandomBlock water = RandomBlock.builder()
                    .block((BlockState) Blocks.field_150358_i.func_176223_P())
                    .height(VariableAmount.baseWithRandomAddition(0, VariableAmount.baseWithRandomAddition(8, 248)))
                    .perChunk(50)
                    .placementTarget(WorldGenConstants.CAVE_LIQUIDS)
                    .build();
            gensettings.getPopulators().add(water);
            RandomBlock lava = RandomBlock.builder()
                    .block((BlockState) Blocks.field_150356_k.func_176223_P())
                    .height(VariableAmount.baseWithRandomAddition(0,
                            VariableAmount.baseWithRandomAddition(8, VariableAmount.baseWithRandomAddition(8, 240))))
                    .perChunk(20)
                    .placementTarget(WorldGenConstants.CAVE_LIQUIDS)
                    .build();
            gensettings.getPopulators().add(lava);
        }

    }

    @Override
    public String getName() {
        return this.biomeName;
    }

    @Override
    public String getId() {
        return this.biomeName;
    }

    @Override
    public double getTemperature() {
        return this.temperature;
    }

    @Override
    public double getHumidity() {
        return this.rainfall;
    }
}
