/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.mixin.core.world.gen.populators;

import com.google.common.base.Objects;
import net.minecraft.block.Block;
import net.minecraft.init.Blocks;
import net.minecraft.util.BlockPos;
import net.minecraft.world.World;
import net.minecraft.world.gen.feature.WorldGenIcePath;
import org.spongepowered.api.util.weighted.VariableAmount;
import org.spongepowered.api.world.Chunk;
import org.spongepowered.api.world.gen.PopulatorType;
import org.spongepowered.api.world.gen.PopulatorTypes;
import org.spongepowered.api.world.gen.populator.IcePath;
import org.spongepowered.asm.mixin.Mixin;
import org.spongepowered.asm.mixin.Overwrite;
import org.spongepowered.asm.mixin.Shadow;
import org.spongepowered.asm.mixin.injection.At;
import org.spongepowered.asm.mixin.injection.Inject;
import org.spongepowered.asm.mixin.injection.callback.CallbackInfo;

import java.util.Random;

@Mixin(WorldGenIcePath.class)
public abstract class MixinWorldGenIcePath implements IcePath {

    private VariableAmount radius;
    private VariableAmount sections;

    @Shadow private Block block;

    @Inject(method = "<init>(I)V", at = @At("RETURN") )
    public void onConstructed(int radius, CallbackInfo ci) {
        this.radius = VariableAmount.baseWithRandomAddition(2, radius > 2 ? radius - 2 : 1);
        this.sections = VariableAmount.fixed(2);
    }

    @Override
    public PopulatorType getType() {
        return PopulatorTypes.ICE_PATH;
    }

    @Override
    public void populate(Chunk chunk, Random random) {
        World world = (World) chunk.getWorld();
        int x;
        int z;
        BlockPos chunkPos = new BlockPos(chunk.getBlockMin().getX(), chunk.getBlockMin().getY(), chunk.getBlockMin().getZ());
        int n = this.sections.getFlooredAmount(random);
        for (int i = 0; i < n; ++i) {
            x = random.nextInt(16) + 8;
            z = random.nextInt(16) + 8;
            generate(world, random, world.func_175645_m(chunkPos.func_177982_a(x, 0, z)));
        }
    }

    /*
     * Author: Deamon - December 12th, 2015
     * 
     * Purpose: Overwritten to replace the path radius with one dependent on
     * our variable amount.
     */
    @Overwrite
    public boolean generate(World worldIn, Random rand, BlockPos position) {
        while (worldIn.func_175623_d(position) && position.func_177956_o() > 2) {
            position = position.func_177977_b();
        }

        if (worldIn.func_180495_p(position).func_177230_c() != Blocks.field_150433_aE) {
            return false;
        }
        // Sponge start
        int i = this.radius.getFlooredAmount(rand);
        // Sponge end
        byte b0 = 1;

        for (int j = position.func_177958_n() - i; j <= position.func_177958_n() + i; ++j) {
            for (int k = position.func_177952_p() - i; k <= position.func_177952_p() + i; ++k) {
                int l = j - position.func_177958_n();
                int i1 = k - position.func_177952_p();

                if (l * l + i1 * i1 <= i * i) {
                    for (int j1 = position.func_177956_o() - b0; j1 <= position.func_177956_o() + b0; ++j1) {
                        BlockPos blockpos1 = new BlockPos(j, j1, k);
                        Block block = worldIn.func_180495_p(blockpos1).func_177230_c();

                        if (block == Blocks.field_150346_d || block == Blocks.field_150433_aE || block == Blocks.field_150432_aD) {
                            worldIn.func_180501_a(blockpos1, this.block.func_176223_P(), 2);
                        }
                    }
                }
            }
        }

        return true;
    }

    @Override
    public VariableAmount getRadius() {
        return this.radius;
    }

    @Override
    public void setRadius(VariableAmount radius) {
        this.radius = radius;
    }

    @Override
    public VariableAmount getSectionsPerChunk() {
        return this.sections;
    }

    @Override
    public void setSectionsPerChunk(VariableAmount sections) {
        this.sections = sections;
    }

    @Override
    public String toString() {
        return Objects.toStringHelper(this)
                .add("Type", "IcePath")
                .add("PerChunk", this.sections)
                .add("Radius", this.radius)
                .toString();
    }

}
