/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.registry.type.item;

import static com.google.common.base.Preconditions.checkNotNull;

import com.google.common.collect.ImmutableList;
import org.spongepowered.api.item.Enchantment;
import org.spongepowered.api.item.Enchantments;
import org.spongepowered.api.registry.AlternateCatalogRegistryModule;
import org.spongepowered.api.registry.util.AdditionalRegistration;
import org.spongepowered.api.registry.util.RegisterCatalog;

import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.Optional;

public final class EnchantmentRegistryModule implements AlternateCatalogRegistryModule<Enchantment> {

    @RegisterCatalog(Enchantments.class)
    private final Map<String, Enchantment> enchantmentMappings = new HashMap<>();

    @Override
    public Optional<Enchantment> getById(String id) {
        checkNotNull(id);
        if (!id.contains(":")) {
            id = "minecraft:" + id; // assume vanilla
        }
        return Optional.ofNullable(this.enchantmentMappings.get(checkNotNull(id).toLowerCase()));
    }

    @Override
    public Collection<Enchantment> getAll() {
        return ImmutableList.copyOf(this.enchantmentMappings.values());
    }

    @Override
    public Map<String, Enchantment> provideCatalogMap() {
        Map<String, Enchantment> newMap = new HashMap<>();
        for (Map.Entry<String, Enchantment> entry : this.enchantmentMappings.entrySet()) {
            newMap.put(entry.getKey().replace("minecraft:", ""), entry.getValue());
        }
        return newMap;
    }

    @Override
    public void registerDefaults() {
        this.enchantmentMappings.put("minecraft:protection", (Enchantment) net.minecraft.enchantment.Enchantment.field_180310_c);
        this.enchantmentMappings.put("minecraft:fire_protection", (Enchantment) net.minecraft.enchantment.Enchantment.field_77329_d);
        this.enchantmentMappings.put("minecraft:feather_falling", (Enchantment) net.minecraft.enchantment.Enchantment.field_180309_e);
        this.enchantmentMappings.put("minecraft:blast_protection", (Enchantment) net.minecraft.enchantment.Enchantment.field_77327_f);
        this.enchantmentMappings.put("minecraft:projectile_protection", (Enchantment) net.minecraft.enchantment.Enchantment.field_180308_g);
        this.enchantmentMappings.put("minecraft:respiration", (Enchantment) net.minecraft.enchantment.Enchantment.field_180317_h);
        this.enchantmentMappings.put("minecraft:aqua_affinity", (Enchantment) net.minecraft.enchantment.Enchantment.field_77341_i);
        this.enchantmentMappings.put("minecraft:thorns", (Enchantment) net.minecraft.enchantment.Enchantment.field_92091_k);
        this.enchantmentMappings.put("minecraft:depth_strider", (Enchantment) net.minecraft.enchantment.Enchantment.field_180316_k);
        this.enchantmentMappings.put("minecraft:sharpness", (Enchantment) net.minecraft.enchantment.Enchantment.field_180314_l);
        this.enchantmentMappings.put("minecraft:smite", (Enchantment) net.minecraft.enchantment.Enchantment.field_180315_m);
        this.enchantmentMappings.put("minecraft:bane_of_arthropods", (Enchantment) net.minecraft.enchantment.Enchantment.field_180312_n);
        this.enchantmentMappings.put("minecraft:knockback", (Enchantment) net.minecraft.enchantment.Enchantment.field_180313_o);
        this.enchantmentMappings.put("minecraft:fire_aspect", (Enchantment) net.minecraft.enchantment.Enchantment.field_77334_n);
        this.enchantmentMappings.put("minecraft:looting", (Enchantment) net.minecraft.enchantment.Enchantment.field_77335_o);
        this.enchantmentMappings.put("minecraft:efficiency", (Enchantment) net.minecraft.enchantment.Enchantment.field_77349_p);
        this.enchantmentMappings.put("minecraft:silk_touch", (Enchantment) net.minecraft.enchantment.Enchantment.field_77348_q);
        this.enchantmentMappings.put("minecraft:unbreaking", (Enchantment) net.minecraft.enchantment.Enchantment.field_77347_r);
        this.enchantmentMappings.put("minecraft:fortune", (Enchantment) net.minecraft.enchantment.Enchantment.field_77346_s);
        this.enchantmentMappings.put("minecraft:power", (Enchantment) net.minecraft.enchantment.Enchantment.field_77345_t);
        this.enchantmentMappings.put("minecraft:punch", (Enchantment) net.minecraft.enchantment.Enchantment.field_77344_u);
        this.enchantmentMappings.put("minecraft:flame", (Enchantment) net.minecraft.enchantment.Enchantment.field_77343_v);
        this.enchantmentMappings.put("minecraft:infinity", (Enchantment) net.minecraft.enchantment.Enchantment.field_77342_w);
        this.enchantmentMappings.put("minecraft:luck_of_the_sea", (Enchantment) net.minecraft.enchantment.Enchantment.field_151370_z);
        this.enchantmentMappings.put("minecraft:lure", (Enchantment) net.minecraft.enchantment.Enchantment.field_151369_A);
    }

    @AdditionalRegistration
    public void registerAdditional() {
        for (net.minecraft.enchantment.Enchantment enchantment : net.minecraft.enchantment.Enchantment.field_180311_a) {
            if (enchantment == null) {
                continue;
            }
            if (!this.enchantmentMappings.containsValue((Enchantment) enchantment)) {
                final String name = enchantment.func_77320_a().replace("enchantment.", "");
                this.enchantmentMappings.put(name.toLowerCase(), (Enchantment) enchantment);
            }
        }

    }
}
