/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.registry.type.world.gen;

import static com.google.common.base.Preconditions.checkNotNull;

import com.google.common.collect.ImmutableList;
import com.google.common.collect.Maps;
import net.minecraft.world.biome.BiomeGenBase;
import org.spongepowered.api.registry.CatalogRegistryModule;
import org.spongepowered.api.registry.util.AdditionalRegistration;
import org.spongepowered.api.registry.util.RegisterCatalog;
import org.spongepowered.api.world.biome.BiomeType;
import org.spongepowered.api.world.biome.BiomeTypes;
import org.spongepowered.common.registry.RegistryHelper;

import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.Map;
import java.util.Optional;

public final class BiomeTypeRegistryModule implements CatalogRegistryModule<BiomeType> {

    @RegisterCatalog(BiomeTypes.class)
    private final Map<String, BiomeType> biomeTypeMappings = Maps.newHashMap();

    private final List<BiomeType> biomeTypes = new ArrayList<>();

    @Override
    public Optional<BiomeType> getById(String id) {
        return Optional.ofNullable(this.biomeTypeMappings.get(checkNotNull(id).toLowerCase()));
    }

    @Override
    public Collection<BiomeType> getAll() {
        return ImmutableList.copyOf(this.biomeTypes);
    }

    @Override
    public void registerDefaults() {
        BiomeGenBase[] biomeArray = BiomeGenBase.func_150565_n();
        for (BiomeGenBase biome : biomeArray) {
            if (biome != null) {
                this.biomeTypes.add((BiomeType) biome);
            }
        }
        this.biomeTypeMappings.put("ocean", (BiomeType) BiomeGenBase.field_76771_b);
        this.biomeTypeMappings.put("plains", (BiomeType) BiomeGenBase.field_76772_c);
        this.biomeTypeMappings.put("desert", (BiomeType) BiomeGenBase.field_76769_d);
        this.biomeTypeMappings.put("extreme_hills", (BiomeType) BiomeGenBase.field_76770_e);
        this.biomeTypeMappings.put("forest", (BiomeType) BiomeGenBase.field_76767_f);
        this.biomeTypeMappings.put("taiga", (BiomeType) BiomeGenBase.field_76768_g);
        this.biomeTypeMappings.put("swampland", (BiomeType) BiomeGenBase.field_76780_h);
        this.biomeTypeMappings.put("river", (BiomeType) BiomeGenBase.field_76781_i);
        this.biomeTypeMappings.put("hell", (BiomeType) BiomeGenBase.field_76778_j);
        this.biomeTypeMappings.put("sky", (BiomeType) BiomeGenBase.field_76779_k);
        this.biomeTypeMappings.put("frozen_ocean", (BiomeType) BiomeGenBase.field_76776_l);
        this.biomeTypeMappings.put("frozen_river", (BiomeType) BiomeGenBase.field_76777_m);
        this.biomeTypeMappings.put("ice_plains", (BiomeType) BiomeGenBase.field_76774_n);
        this.biomeTypeMappings.put("ice_mountains", (BiomeType) BiomeGenBase.field_76775_o);
        this.biomeTypeMappings.put("mushroom_island", (BiomeType) BiomeGenBase.field_76789_p);
        this.biomeTypeMappings.put("mushroom_island_shore", (BiomeType) BiomeGenBase.field_76788_q);
        this.biomeTypeMappings.put("beach", (BiomeType) BiomeGenBase.field_76787_r);
        this.biomeTypeMappings.put("desert_hills", (BiomeType) BiomeGenBase.field_76786_s);
        this.biomeTypeMappings.put("forest_hills", (BiomeType) BiomeGenBase.field_76785_t);
        this.biomeTypeMappings.put("taiga_hills", (BiomeType) BiomeGenBase.field_76784_u);
        this.biomeTypeMappings.put("extreme_hills_edge", (BiomeType) BiomeGenBase.field_76783_v);
        this.biomeTypeMappings.put("jungle", (BiomeType) BiomeGenBase.field_76782_w);
        this.biomeTypeMappings.put("jungle_hills", (BiomeType) BiomeGenBase.field_76792_x);
        this.biomeTypeMappings.put("jungle_edge", (BiomeType) BiomeGenBase.field_150574_L);
        this.biomeTypeMappings.put("deep_ocean", (BiomeType) BiomeGenBase.field_150575_M);
        this.biomeTypeMappings.put("stone_beach", (BiomeType) BiomeGenBase.field_150576_N);
        this.biomeTypeMappings.put("cold_beach", (BiomeType) BiomeGenBase.field_150577_O);
        this.biomeTypeMappings.put("birch_forest", (BiomeType) BiomeGenBase.field_150583_P);
        this.biomeTypeMappings.put("birch_forest_hills", (BiomeType) BiomeGenBase.field_150582_Q);
        this.biomeTypeMappings.put("roofed_forest", (BiomeType) BiomeGenBase.field_150585_R);
        this.biomeTypeMappings.put("cold_taiga", (BiomeType) BiomeGenBase.field_150584_S);
        this.biomeTypeMappings.put("cold_taiga_hills", (BiomeType) BiomeGenBase.field_150579_T);
        this.biomeTypeMappings.put("mega_taiga", (BiomeType) BiomeGenBase.field_150578_U);
        this.biomeTypeMappings.put("mega_taiga_hills", (BiomeType) BiomeGenBase.field_150581_V);
        this.biomeTypeMappings.put("extreme_hills_plus", (BiomeType) BiomeGenBase.field_150580_W);
        this.biomeTypeMappings.put("savanna", (BiomeType) BiomeGenBase.field_150588_X);
        this.biomeTypeMappings.put("savanna_plateau", (BiomeType) BiomeGenBase.field_150587_Y);
        this.biomeTypeMappings.put("mesa", (BiomeType) BiomeGenBase.field_150589_Z);
        this.biomeTypeMappings.put("mesa_plateau_forest", (BiomeType) BiomeGenBase.field_150607_aa);
        this.biomeTypeMappings.put("mesa_plateau", (BiomeType) BiomeGenBase.field_150608_ab);
        this.biomeTypeMappings.put("sunflower_plains", (BiomeType) biomeArray[BiomeGenBase.field_76772_c.field_76756_M + 128]);
        this.biomeTypeMappings.put("desert_mountains", (BiomeType) biomeArray[BiomeGenBase.field_76769_d.field_76756_M + 128]);
        this.biomeTypeMappings.put("flower_forest", (BiomeType) biomeArray[BiomeGenBase.field_76767_f.field_76756_M + 128]);
        this.biomeTypeMappings.put("taiga_mountains", (BiomeType) biomeArray[BiomeGenBase.field_76768_g.field_76756_M + 128]);
        this.biomeTypeMappings.put("swampland_mountains", (BiomeType) biomeArray[BiomeGenBase.field_76780_h.field_76756_M + 128]);
        this.biomeTypeMappings.put("ice_plains_spikes", (BiomeType) biomeArray[BiomeGenBase.field_76774_n.field_76756_M + 128]);
        this.biomeTypeMappings.put("jungle_mountains", (BiomeType) biomeArray[BiomeGenBase.field_76782_w.field_76756_M + 128]);
        this.biomeTypeMappings.put("jungle_edge_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150574_L.field_76756_M + 128]);
        this.biomeTypeMappings.put("cold_taiga_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150584_S.field_76756_M + 128]);
        this.biomeTypeMappings.put("savanna_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150588_X.field_76756_M + 128]);
        this.biomeTypeMappings.put("savanna_plateau_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150587_Y.field_76756_M + 128]);
        this.biomeTypeMappings.put("mesa_bryce", (BiomeType) biomeArray[BiomeGenBase.field_150589_Z.field_76756_M + 128]);
        this.biomeTypeMappings.put("mesa_plateau_forest_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150607_aa.field_76756_M + 128]);
        this.biomeTypeMappings.put("mesa_plateau_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150608_ab.field_76756_M + 128]);
        this.biomeTypeMappings.put("birch_forest_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150583_P.field_76756_M + 128]);
        this.biomeTypeMappings.put("birch_forest_hills_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150582_Q.field_76756_M + 128]);
        this.biomeTypeMappings.put("roofed_forest_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150585_R.field_76756_M + 128]);
        this.biomeTypeMappings.put("mega_spruce_taiga", (BiomeType) biomeArray[BiomeGenBase.field_150578_U.field_76756_M + 128]);
        this.biomeTypeMappings.put("extreme_hills_mountains", (BiomeType) biomeArray[BiomeGenBase.field_76770_e.field_76756_M + 128]);
        this.biomeTypeMappings.put("extreme_hills_plus_mountains", (BiomeType) biomeArray[BiomeGenBase.field_150580_W.field_76756_M + 128]);
        this.biomeTypeMappings.put("mega_spruce_taiga_hills", (BiomeType) biomeArray[BiomeGenBase.field_150581_V.field_76756_M + 128]);
    }

    @AdditionalRegistration
    public void registerAdditional() {
        BiomeGenBase[] biomeArray = BiomeGenBase.func_150565_n();
        for (BiomeGenBase biome : biomeArray) {
            if (biome != null && !this.biomeTypes.contains((BiomeType) biome)) {
                this.biomeTypes.add((BiomeType) biome);
                this.biomeTypeMappings.put(biome.field_76791_y.toLowerCase(), (BiomeType) biome);
            }
        }
        // Re-map fields in case mods have changed vanilla world types
        RegistryHelper.mapFields(BiomeTypes.class, this.biomeTypeMappings);
    }
}
