/*
 * This file is part of Sponge, licensed under the MIT License (MIT).
 *
 * Copyright (c) SpongePowered <https://www.spongepowered.org>
 * Copyright (c) contributors
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */
package org.spongepowered.common.world.gen.populators;

import com.flowpowered.math.vector.Vector3i;
import com.google.common.base.Objects;
import net.minecraft.util.BlockPos;
import org.spongepowered.api.util.weighted.VariableAmount;
import org.spongepowered.api.util.weighted.WeightedTable;
import org.spongepowered.api.world.Chunk;
import org.spongepowered.api.world.Location;
import org.spongepowered.api.world.gen.PopulatorObject;
import org.spongepowered.api.world.gen.PopulatorType;
import org.spongepowered.api.world.gen.PopulatorTypes;
import org.spongepowered.api.world.gen.populator.Forest;
import org.spongepowered.common.util.VecHelper;

import java.util.List;
import java.util.Optional;
import java.util.Random;
import java.util.function.Function;

import javax.annotation.Nullable;

public class ForestPopulator implements Forest {

    private VariableAmount count;
    private WeightedTable<PopulatorObject> types;
    private Function<Location<Chunk>, PopulatorObject> override = null;

    public ForestPopulator() {
        this.count = VariableAmount.fixed(10);
        this.types = new WeightedTable<>();
    }

    @Override
    public PopulatorType getType() {
        return PopulatorTypes.FOREST;
    }

    @Override
    public void populate(Chunk chunk, Random random) {
        Vector3i min = chunk.getBlockMin();
        int n = this.count.getFlooredAmount(random);
        int x;
        int z;
        BlockPos pos;

        List<PopulatorObject> result;
        PopulatorObject type;
        for (int i = 0; i < n; i++) {
            x = random.nextInt(16) + 8;
            z = random.nextInt(16) + 8;
            pos = ((net.minecraft.world.World) chunk.getWorld()).func_175672_r(new BlockPos(min.getX() + x, min.getY(), min.getZ() + z));
            if (this.override != null) {
                Location<Chunk> pos2 = new Location<>(chunk, VecHelper.toVector(pos));
                type = this.override.apply(pos2);
            } else {
                result = this.types.get(random);
                if (result.isEmpty()) {
                    continue;
                }
                type = result.get(0);
            }
            if (type.canPlaceAt(chunk.getWorld(), pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p())) {
                type.placeObject(chunk.getWorld(), random, pos.func_177958_n(), pos.func_177956_o(), pos.func_177952_p());
            }
        }
    }

    @Override
    public VariableAmount getTreesPerChunk() {
        return this.count;
    }

    @Override
    public void setTreesPerChunk(VariableAmount count) {
        this.count = count;
    }

    @Override
    public WeightedTable<PopulatorObject> getTypes() {
        return this.types;
    }

    @Override
    public Optional<Function<Location<Chunk>, PopulatorObject>> getSupplierOverride() {
        return Optional.ofNullable(this.override);
    }

    @Override
    public void setSupplierOverride(@Nullable Function<Location<Chunk>, PopulatorObject> override) {
        this.override = override;
    }

    @Override
    public String toString() {
        return Objects.toStringHelper(this)
                .add("count", this.count)
                .add("types", this.types)
                .toString();
    }

}
